@@ -1,6 +1,5 @@
 module TextBuilderDev
-  ( -- *
-    TextBuilder,
+  ( TextBuilder,
 
     -- * Accessors
     buildText,
@@ -18,6 +17,7 @@ module TextBuilderDev
     -- ** Builder manipulators
     force,
     intercalate,
+    intercalateMap,
     padFromLeft,
     padFromRight,
 
@@ -45,6 +45,7 @@ module TextBuilderDev
     -- *** Decimal
     decimal,
     unsignedDecimal,
+    fixedUnsignedDecimal,
     thousandSeparatedDecimal,
     thousandSeparatedUnsignedDecimal,
     dataSizeInBytesInDecimal,
@@ -52,6 +53,7 @@ module TextBuilderDev
     -- *** Binary
     unsignedBinary,
     unsignedPaddedBinary,
+    finiteBitsUnsignedBinary,
 
     -- *** Hexadecimal
     hexadecimal,
@@ -66,6 +68,7 @@ module TextBuilderDev
     doublePercent,
 
     -- ** Time
+    utcTimeInIso8601,
     utcTimestampInIso8601,
     intervalInSeconds,
 
@@ -77,22 +80,15 @@ where
 import qualified Data.ByteString as ByteString
 import qualified Data.List.Split as Split
 import qualified Data.Text as Text
-import qualified Data.Text.Array as B
-import qualified Data.Text.Encoding as E
-import qualified Data.Text.Encoding.Error as E
 import qualified Data.Text.IO as Text
-import qualified Data.Text.Internal as C
-import qualified DeferredFolds.Unfoldr as Unfoldr
-import TextBuilderDev.Prelude hiding (intercalate, length, null)
-#if MIN_VERSION_text(2,0,0)
-import qualified TextBuilderDev.UTF8 as D
-#else
-import qualified TextBuilderDev.UTF16 as D
-#endif
 import qualified Data.Text.Lazy as TextLazy
 import qualified Data.Text.Lazy.Builder as TextLazyBuilder
+import qualified DeferredFolds.Unfoldr as Unfoldr
+import qualified Test.QuickCheck.Gen as QcGen
+import qualified TextBuilderDev.Allocator as Allocator
+import TextBuilderDev.Prelude hiding (intercalate, length, null)
 
--- *
+-- * --
 
 -- |
 -- Evidence that there exists an unambiguous way to convert
@@ -142,34 +138,27 @@ instance IsomorphicToTextBuilder TextLazyBuilder.Builder where
   toTextBuilder = text . TextLazy.toStrict . TextLazyBuilder.toLazyText
   fromTextBuilder = TextLazyBuilder.fromText . buildText
 
--- *
+-- * --
 
 -- |
 -- Specification of how to efficiently construct strict 'Text'.
 -- Provides instances of 'Semigroup' and 'Monoid', which have complexity of /O(1)/.
 data TextBuilder
-  = TextBuilder !Action !Int !Int
-
-newtype Action
-  = Action (forall s. B.MArray s -> Int -> ST s ())
+  = TextBuilder
+      {-# UNPACK #-} !Allocator.Allocator
+      {-# UNPACK #-} !Int
 
 instance Semigroup TextBuilder where
-  (<>) (TextBuilder (Action action1) arraySize1 charsAmount1) (TextBuilder (Action action2) arraySize2 charsAmount2) =
-    TextBuilder action arraySize charsAmount
-    where
-      action =
-        Action $ \array offset -> do
-          action1 array offset
-          action2 array (offset + arraySize1)
-      arraySize =
-        arraySize1 + arraySize2
-      charsAmount =
-        charsAmount1 + charsAmount2
+  (<>) (TextBuilder allocator1 sizeInChars1) (TextBuilder allocator2 sizeInChars2) =
+    TextBuilder
+      (allocator1 <> allocator2)
+      (sizeInChars1 + sizeInChars2)
+  stimes n (TextBuilder allocator size) =
+    TextBuilder (stimes n allocator) (size * fromIntegral n)
 
 instance Monoid TextBuilder where
   {-# INLINE mempty #-}
-  mempty =
-    TextBuilder (Action (\_ _ -> return ())) 0 0
+  mempty = TextBuilder mempty 0
 
 instance IsString TextBuilder where
   fromString = string
@@ -180,44 +169,100 @@ instance Show TextBuilder where
 instance Eq TextBuilder where
   (==) = on (==) buildText
 
+instance Arbitrary TextBuilder where
+  arbitrary =
+    QcGen.oneof
+      [ QcGen.scale (flip div 2)
+          $ QcGen.oneof
+            [ (<>) <$> arbitrary <*> arbitrary,
+              sconcat <$> arbitrary,
+              stimes <$> arbitrary @Word8 <*> arbitrary,
+              pure mempty,
+              mconcat <$> arbitrary
+            ],
+        text <$> arbitrary,
+        lazyText <$> arbitrary,
+        string <$> arbitrary,
+        asciiByteString . ByteString.filter (< 128) <$> arbitrary,
+        hexData <$> arbitrary,
+        char <$> arbitrary,
+        decimal @Integer <$> arbitrary,
+        unsignedDecimal @Natural <$> arbitrary,
+        thousandSeparatedDecimal @Integer <$> arbitrary <*> arbitrary,
+        thousandSeparatedUnsignedDecimal @Natural <$> arbitrary <*> arbitrary,
+        dataSizeInBytesInDecimal @Natural <$> arbitrary <*> arbitrary,
+        unsignedBinary @Natural <$> arbitrary,
+        unsignedPaddedBinary @Word <$> arbitrary,
+        finiteBitsUnsignedBinary @Word <$> arbitrary,
+        hexadecimal @Integer <$> arbitrary,
+        unsignedHexadecimal @Natural <$> arbitrary,
+        decimalDigit <$> QcGen.choose (0, 9),
+        hexadecimalDigit <$> QcGen.choose (0, 15),
+        fixedDouble <$> QcGen.choose (0, 19) <*> arbitrary,
+        doublePercent <$> QcGen.choose (0, 19) <*> arbitrary,
+        utcTimestampInIso8601 <$> arbitrary <*> arbitrary <*> arbitrary <*> arbitrary <*> arbitrary <*> arbitrary,
+        intervalInSeconds @Double <$> arbitrary
+      ]
+
+instance IsomorphicTo TextBuilder TextBuilder where
+  to = id
+
+instance IsomorphicTo TextBuilder String where
+  to = TextBuilderDev.string
+
+instance IsomorphicTo TextBuilder Text where
+  to = TextBuilderDev.text
+
+instance IsomorphicTo TextBuilder TextLazy.Text where
+  to = TextBuilderDev.lazyText
+
+instance IsomorphicTo TextBuilder TextLazyBuilder.Builder where
+  to = to . to @TextLazy.Text
+
+instance IsomorphicTo String TextBuilder where
+  to = to . to @Text
+
+instance IsomorphicTo Text TextBuilder where
+  to = TextBuilderDev.buildText
+
+instance IsomorphicTo TextLazy.Text TextBuilder where
+  to = to . to @Text
+
+instance IsomorphicTo TextLazyBuilder.Builder TextBuilder where
+  to = to . to @Text
+
 -- * Accessors
 
--- | Get the amount of characters
+-- | Get the amount of characters.
 {-# INLINE length #-}
 length :: TextBuilder -> Int
-length (TextBuilder _ _ x) = x
+length (TextBuilder _ x) = x
 
--- | Check whether the builder is empty
+-- | Check whether the builder is empty.
 {-# INLINE null #-}
 null :: TextBuilder -> Bool
 null = (== 0) . length
 
--- | Execute a builder producing a strict text
+-- | Execute a builder producing a strict text.
 buildText :: TextBuilder -> Text
-buildText (TextBuilder (Action action) arraySize _) =
-  C.text array 0 arraySize
-  where
-    array =
-      runST $ do
-        array <- B.new arraySize
-        action array 0
-        B.unsafeFreeze array
+buildText (TextBuilder allocator _) =
+  Allocator.allocate allocator
 
 -- ** Output IO
 
--- | Put builder, to stdout
+-- | Put builder, to stdout.
 putToStdOut :: TextBuilder -> IO ()
 putToStdOut = Text.hPutStr stdout . buildText
 
--- | Put builder, to stderr
+-- | Put builder, to stderr.
 putToStdErr :: TextBuilder -> IO ()
 putToStdErr = Text.hPutStr stderr . buildText
 
--- | Put builder, followed by a line, to stdout
+-- | Put builder, followed by a line, to stdout.
 putLnToStdOut :: TextBuilder -> IO ()
 putLnToStdOut = Text.hPutStrLn stdout . buildText
 
--- | Put builder, followed by a line, to stderr
+-- | Put builder, followed by a line, to stderr.
 putLnToStdErr :: TextBuilder -> IO ()
 putLnToStdErr = Text.hPutStrLn stderr . buildText
 
@@ -233,187 +278,138 @@ putLnToStdErr = Text.hPutStrLn stderr . buildText
 force :: TextBuilder -> TextBuilder
 force = text . buildText
 
--- | Unicode character
+-- | Unicode character.
 {-# INLINE char #-}
 char :: Char -> TextBuilder
-char x =
-  unicodeCodePoint (ord x)
+char = unicodeCodePoint . ord
 
-#if MIN_VERSION_text(2,0,0)
-
--- | Unicode code point
+-- | Unicode code point.
 {-# INLINE unicodeCodePoint #-}
 unicodeCodePoint :: Int -> TextBuilder
-unicodeCodePoint x =
-  D.unicodeCodePoint x utf8CodeUnits1 utf8CodeUnits2 utf8CodeUnits3 utf8CodeUnits4
+unicodeCodePoint a =
+  TextBuilder (Allocator.unicodeCodePoint a) 1
 
-{-# INLINEABLE utf8CodeUnits1 #-}
-utf8CodeUnits1 :: Word8 -> TextBuilder
-utf8CodeUnits1 unit1 = TextBuilder action 1 1
-  where action = Action $ \array offset -> B.unsafeWrite array offset unit1
-
-{-# INLINEABLE utf8CodeUnits2 #-}
-utf8CodeUnits2 :: Word8 -> Word8 -> TextBuilder
-utf8CodeUnits2 unit1 unit2 = TextBuilder action 2 1
-  where action = Action $ \array offset -> do
-          B.unsafeWrite array (offset + 0) unit1
-          B.unsafeWrite array (offset + 1) unit2
-
-{-# INLINEABLE utf8CodeUnits3 #-}
-utf8CodeUnits3 :: Word8 -> Word8 -> Word8 -> TextBuilder
-utf8CodeUnits3 unit1 unit2 unit3 = TextBuilder action 3 1
-  where action = Action $ \array offset -> do
-          B.unsafeWrite array (offset + 0) unit1
-          B.unsafeWrite array (offset + 1) unit2
-          B.unsafeWrite array (offset + 2) unit3
-
-{-# INLINEABLE utf8CodeUnits4 #-}
-utf8CodeUnits4 :: Word8 -> Word8 -> Word8 -> Word8 -> TextBuilder
-utf8CodeUnits4 unit1 unit2 unit3 unit4 = TextBuilder action 4 1
-  where action = Action $ \array offset -> do
-          B.unsafeWrite array (offset + 0) unit1
-          B.unsafeWrite array (offset + 1) unit2
-          B.unsafeWrite array (offset + 2) unit3
-          B.unsafeWrite array (offset + 3) unit4
-
-{-# INLINE utf16CodeUnits1 #-}
-utf16CodeUnits1 :: Word16 -> TextBuilder
-utf16CodeUnits1 = unicodeCodePoint . fromIntegral
-
-{-# INLINE utf16CodeUnits2 #-}
-utf16CodeUnits2 :: Word16 -> Word16 -> TextBuilder
-utf16CodeUnits2 unit1 unit2 = unicodeCodePoint cp
-  where
-    cp = (((fromIntegral unit1 .&. 0x3FF) `shiftL` 10) .|. (fromIntegral unit2 .&. 0x3FF)) + 0x10000
-
-#else
-
--- | Unicode code point
-{-# INLINE unicodeCodePoint #-}
-unicodeCodePoint :: Int -> TextBuilder
-unicodeCodePoint x =
-  D.unicodeCodePoint x utf16CodeUnits1 utf16CodeUnits2
-
--- | Single code-unit UTF-16 character
+-- | Single code-unit UTF-16 character.
 {-# INLINEABLE utf16CodeUnits1 #-}
 utf16CodeUnits1 :: Word16 -> TextBuilder
-utf16CodeUnits1 unit =
-  TextBuilder action 1 1
-  where
-    action =
-      Action $ \array offset -> B.unsafeWrite array offset unit
+utf16CodeUnits1 a =
+  TextBuilder (Allocator.utf16CodeUnits1 a) 1
 
--- | Double code-unit UTF-16 character
+-- | Double code-unit UTF-16 character.
 {-# INLINEABLE utf16CodeUnits2 #-}
 utf16CodeUnits2 :: Word16 -> Word16 -> TextBuilder
-utf16CodeUnits2 unit1 unit2 =
-  TextBuilder action 2 1
-  where
-    action =
-      Action $ \array offset -> do
-        B.unsafeWrite array offset unit1
-        B.unsafeWrite array (succ offset) unit2
+utf16CodeUnits2 a b =
+  TextBuilder (Allocator.utf16CodeUnits2 a b) 1
 
--- | Single code-unit UTF-8 character
+-- | Single code-unit UTF-8 character.
 {-# INLINE utf8CodeUnits1 #-}
 utf8CodeUnits1 :: Word8 -> TextBuilder
-utf8CodeUnits1 unit1 =
-  D.utf8CodeUnits1 unit1 utf16CodeUnits1 utf16CodeUnits2
+utf8CodeUnits1 a =
+  TextBuilder (Allocator.utf8CodeUnits1 a) 1
 
--- | Double code-unit UTF-8 character
+-- | Double code-unit UTF-8 character.
 {-# INLINE utf8CodeUnits2 #-}
 utf8CodeUnits2 :: Word8 -> Word8 -> TextBuilder
-utf8CodeUnits2 unit1 unit2 =
-  D.utf8CodeUnits2 unit1 unit2 utf16CodeUnits1 utf16CodeUnits2
+utf8CodeUnits2 a b =
+  TextBuilder (Allocator.utf8CodeUnits2 a b) 1
 
--- | Triple code-unit UTF-8 character
+-- | Triple code-unit UTF-8 character.
 {-# INLINE utf8CodeUnits3 #-}
 utf8CodeUnits3 :: Word8 -> Word8 -> Word8 -> TextBuilder
-utf8CodeUnits3 unit1 unit2 unit3 =
-  D.utf8CodeUnits3 unit1 unit2 unit3 utf16CodeUnits1 utf16CodeUnits2
+utf8CodeUnits3 a b c =
+  TextBuilder (Allocator.utf8CodeUnits3 a b c) 1
 
--- | UTF-8 character out of 4 code units
+-- | UTF-8 character out of 4 code units.
 {-# INLINE utf8CodeUnits4 #-}
 utf8CodeUnits4 :: Word8 -> Word8 -> Word8 -> Word8 -> TextBuilder
-utf8CodeUnits4 unit1 unit2 unit3 unit4 =
-  D.utf8CodeUnits4 unit1 unit2 unit3 unit4 utf16CodeUnits1 utf16CodeUnits2
+utf8CodeUnits4 a b c d =
+  TextBuilder (Allocator.utf8CodeUnits4 a b c d) 1
 
-#endif
-
--- | ASCII byte string
+-- | ASCII byte string.
+--
+-- It's your responsibility to ensure that the bytes are in proper range,
+-- otherwise the produced text will be broken.
 {-# INLINEABLE asciiByteString #-}
 asciiByteString :: ByteString -> TextBuilder
 asciiByteString byteString =
-  TextBuilder action length length
-  where
-    length = ByteString.length byteString
-    action =
-      Action $ \array ->
-        let step byte next index = do
-              B.unsafeWrite array index (fromIntegral byte)
-              next (succ index)
-         in ByteString.foldr step (const (return ())) byteString
+  TextBuilder
+    (Allocator.asciiByteString byteString)
+    (ByteString.length byteString)
 
--- | Strict text
+-- | Strict text.
 {-# INLINEABLE text #-}
 text :: Text -> TextBuilder
-text text@(C.Text array offset length) =
-  TextBuilder action length (Text.length text)
-  where
-    action =
-      Action $ \builderArray builderOffset -> do
-#if MIN_VERSION_text(2,0,0)
-        B.copyI length builderArray builderOffset array offset
-#else
-        B.copyI builderArray builderOffset array offset (builderOffset + length)
-#endif
+text text =
+  TextBuilder (Allocator.text text) (Text.length text)
 
+-- | Lazy text.
 {-# INLINE lazyText #-}
 lazyText :: TextLazy.Text -> TextBuilder
 lazyText =
   TextLazy.foldrChunks (mappend . text) mempty
 
--- | String
+-- | String.
 {-# INLINE string #-}
 string :: String -> TextBuilder
 string =
   foldMap char
 
--- | Decimal representation of an integral value
+-- | Decimal representation of an integral value.
 {-# INLINEABLE decimal #-}
-decimal :: Integral a => a -> TextBuilder
+decimal :: (Integral a) => a -> TextBuilder
 decimal i =
   if i >= 0
     then unsignedDecimal i
     else unicodeCodePoint 45 <> unsignedDecimal (negate i)
 
--- | Decimal representation of an unsigned integral value
+-- | Decimal representation of an unsigned integral value.
 {-# INLINEABLE unsignedDecimal #-}
-unsignedDecimal :: Integral a => a -> TextBuilder
+unsignedDecimal :: (Integral a) => a -> TextBuilder
 unsignedDecimal =
-  foldMap decimalDigit . Unfoldr.decimalDigits
+  foldMap (decimalDigit . fromIntegral) . Unfoldr.decimalDigits
 
--- | Decimal representation of an integral value with thousands separated by the specified character
+fixedUnsignedDecimal :: (Integral a) => Int -> a -> TextBuilder
+fixedUnsignedDecimal size val =
+  TextBuilder (Allocator.fixedUnsignedDecimal size val) size
+
+-- | Decimal representation of an integral value with thousands separated by the specified character.
 {-# INLINEABLE thousandSeparatedDecimal #-}
-thousandSeparatedDecimal :: Integral a => Char -> a -> TextBuilder
+thousandSeparatedDecimal :: (Integral a) => Char -> a -> TextBuilder
 thousandSeparatedDecimal separatorChar a =
   if a >= 0
     then thousandSeparatedUnsignedDecimal separatorChar a
     else unicodeCodePoint 45 <> thousandSeparatedUnsignedDecimal separatorChar (negate a)
 
--- | Decimal representation of an unsigned integral value with thousands separated by the specified character
+-- | Decimal representation of an unsigned integral value with thousands separated by the specified character.
 {-# INLINEABLE thousandSeparatedUnsignedDecimal #-}
-thousandSeparatedUnsignedDecimal :: Integral a => Char -> a -> TextBuilder
-thousandSeparatedUnsignedDecimal separatorChar a =
-  fold $ do
-    (index, digit) <- Unfoldr.zipWithReverseIndex $ Unfoldr.decimalDigits a
-    if mod index 3 == 0 && index /= 0
-      then return (decimalDigit digit <> char separatorChar)
-      else return (decimalDigit digit)
+thousandSeparatedUnsignedDecimal :: (Integral a) => Char -> a -> TextBuilder
+thousandSeparatedUnsignedDecimal separatorChar =
+  processRightmostDigit
+  where
+    processRightmostDigit value =
+      case divMod value 10 of
+        (value, digit) ->
+          processAnotherDigit [decimalDigit (fromIntegral digit)] 1 value
+    processAnotherDigit builders index value =
+      if value == 0
+        then mconcat builders
+        else case divMod value 10 of
+          (value, digit) ->
+            if mod index 3 == 0
+              then
+                processAnotherDigit
+                  (decimalDigit (fromIntegral digit) : char separatorChar : builders)
+                  (succ index)
+                  value
+              else
+                processAnotherDigit
+                  (decimalDigit (fromIntegral digit) : builders)
+                  (succ index)
+                  value
 
 -- | Data size in decimal notation over amount of bytes.
 {-# INLINEABLE dataSizeInBytesInDecimal #-}
-dataSizeInBytesInDecimal :: Integral a => Char -> a -> TextBuilder
+dataSizeInBytesInDecimal :: (Integral a) => Char -> a -> TextBuilder
 dataSizeInBytesInDecimal separatorChar amount =
   if amount < 1000
     then unsignedDecimal amount <> "B"
@@ -440,69 +436,89 @@ dataSizeInBytesInDecimal separatorChar amount =
                                 then dividedDecimal separatorChar 100000000000000000000 amount <> "ZB"
                                 else dividedDecimal separatorChar 100000000000000000000000 amount <> "YB"
 
-dividedDecimal :: Integral a => Char -> a -> a -> TextBuilder
+dividedDecimal :: (Integral a) => Char -> a -> a -> TextBuilder
 dividedDecimal separatorChar divisor n =
   let byDivisor = div n divisor
       byExtraTen = div byDivisor 10
       remainder = byDivisor - byExtraTen * 10
    in if remainder == 0 || byExtraTen >= 10
         then thousandSeparatedDecimal separatorChar byExtraTen
-        else thousandSeparatedDecimal separatorChar byExtraTen <> "." <> decimalDigit remainder
+        else thousandSeparatedDecimal separatorChar byExtraTen <> "." <> decimalDigit (fromIntegral remainder)
 
--- | Unsigned binary number
+-- | Unsigned binary number.
 {-# INLINE unsignedBinary #-}
-unsignedBinary :: Integral a => a -> TextBuilder
+unsignedBinary :: (Integral a) => a -> TextBuilder
 unsignedBinary =
-  foldMap decimalDigit . Unfoldr.binaryDigits
+  foldMap (decimalDigit . fromIntegral) . Unfoldr.binaryDigits
 
--- | Unsigned binary number
+-- | A less general but faster alternative to 'unsignedBinary'.
+finiteBitsUnsignedBinary :: (FiniteBits a) => a -> TextBuilder
+finiteBitsUnsignedBinary a =
+  TextBuilder allocator size
+  where
+    allocator = Allocator.finiteBitsUnsignedBinary a
+    size = Allocator.sizeBound allocator
+
+-- | Unsigned binary number.
 {-# INLINE unsignedPaddedBinary #-}
 unsignedPaddedBinary :: (Integral a, FiniteBits a) => a -> TextBuilder
 unsignedPaddedBinary a =
-  padFromLeft (finiteBitSize a) '0' $ foldMap decimalDigit $ Unfoldr.binaryDigits a
+  padFromLeft (finiteBitSize a) '0' $ foldMap (decimalDigit . fromIntegral) $ Unfoldr.binaryDigits a
 
--- | Hexadecimal representation of an integral value
+-- | Hexadecimal representation of an integral value.
 {-# INLINE hexadecimal #-}
-hexadecimal :: Integral a => a -> TextBuilder
+hexadecimal :: (Integral a) => a -> TextBuilder
 hexadecimal i =
   if i >= 0
     then unsignedHexadecimal i
     else unicodeCodePoint 45 <> unsignedHexadecimal (negate i)
 
--- | Unsigned hexadecimal representation of an integral value
+-- | Unsigned hexadecimal representation of an integral value.
 {-# INLINE unsignedHexadecimal #-}
-unsignedHexadecimal :: Integral a => a -> TextBuilder
+unsignedHexadecimal :: (Integral a) => a -> TextBuilder
 unsignedHexadecimal =
-  foldMap hexadecimalDigit . Unfoldr.hexadecimalDigits
+  foldMap (hexadecimalDigit . fromIntegral) . Unfoldr.hexadecimalDigits
 
--- | Decimal digit
+-- | Decimal digit.
 {-# INLINE decimalDigit #-}
-decimalDigit :: Integral a => a -> TextBuilder
+decimalDigit :: Int -> TextBuilder
 decimalDigit n =
-  unicodeCodePoint (fromIntegral n + 48)
+  unicodeCodePoint (n + 48)
 
--- | Hexadecimal digit
+-- | Hexadecimal digit.
 {-# INLINE hexadecimalDigit #-}
-hexadecimalDigit :: Integral a => a -> TextBuilder
+hexadecimalDigit :: Int -> TextBuilder
 hexadecimalDigit n =
   if n <= 9
-    then unicodeCodePoint (fromIntegral n + 48)
-    else unicodeCodePoint (fromIntegral n + 87)
+    then unicodeCodePoint (n + 48)
+    else unicodeCodePoint (n + 87)
 
--- | Intercalate builders
+-- | Intercalate builders.
 {-# INLINE intercalate #-}
-intercalate :: Foldable foldable => TextBuilder -> foldable TextBuilder -> TextBuilder
+intercalate :: (Foldable f) => TextBuilder -> f TextBuilder -> TextBuilder
 intercalate separator = extract . foldl' step init
   where
     init = Product2 False mempty
     step (Product2 isNotFirst builder) element =
-      Product2 True $
-        if isNotFirst
+      Product2 True
+        $ if isNotFirst
           then builder <> separator <> element
           else element
     extract (Product2 _ builder) = builder
 
--- | Pad a builder from the left side to the specified length with the specified character
+-- | Intercalate projecting values to builder.
+{-# INLINE intercalateMap #-}
+intercalateMap :: (Foldable f) => TextBuilder -> (a -> TextBuilder) -> f a -> TextBuilder
+intercalateMap separator mapper = extract . foldl' step init
+  where
+    init = Nothing
+    step acc element =
+      Just $ case acc of
+        Nothing -> mapper element
+        Just acc -> acc <> separator <> mapper element
+    extract = fromMaybe mempty
+
+-- | Pad a builder from the left side to the specified length with the specified character.
 {-# INLINEABLE padFromLeft #-}
 padFromLeft :: Int -> Char -> TextBuilder -> TextBuilder
 padFromLeft paddedLength paddingChar builder =
@@ -511,7 +527,7 @@ padFromLeft paddedLength paddingChar builder =
         then builder
         else foldMap char (replicate (paddedLength - builderLength) paddingChar) <> builder
 
--- | Pad a builder from the right side to the specified length with the specified character
+-- | Pad a builder from the right side to the specified length with the specified character.
 {-# INLINEABLE padFromRight #-}
 padFromRight :: Int -> Char -> TextBuilder -> TextBuilder
 padFromRight paddedLength paddingChar builder =
@@ -520,6 +536,14 @@ padFromRight paddedLength paddingChar builder =
         then builder
         else builder <> foldMap char (replicate (paddedLength - builderLength) paddingChar)
 
+utcTimeInIso8601 :: UTCTime -> TextBuilder
+utcTimeInIso8601 UTCTime {..} =
+  let (year, month, day) = toGregorian utctDay
+      daySeconds = round utctDayTime
+      (dayMinutes, second) = divMod daySeconds 60
+      (hour, minute) = divMod dayMinutes 60
+   in utcTimestampInIso8601 (fromIntegral year) month day hour minute second
+
 -- |
 -- General template for formatting date values according to the ISO8601 standard.
 -- The format is the following:
@@ -543,17 +567,17 @@ utcTimestampInIso8601 ::
   TextBuilder
 utcTimestampInIso8601 y mo d h mi s =
   mconcat
-    [ padFromLeft 4 '0' $ decimal y,
+    [ fixedUnsignedDecimal 4 y,
       "-",
-      padFromLeft 2 '0' $ decimal mo,
+      fixedUnsignedDecimal 2 mo,
       "-",
-      padFromLeft 2 '0' $ decimal d,
+      fixedUnsignedDecimal 2 d,
       "T",
-      padFromLeft 2 '0' $ decimal h,
+      fixedUnsignedDecimal 2 h,
       ":",
-      padFromLeft 2 '0' $ decimal mi,
+      fixedUnsignedDecimal 2 mi,
       ":",
-      padFromLeft 2 '0' $ decimal s,
+      fixedUnsignedDecimal 2 s,
       "Z"
     ]
 
@@ -561,19 +585,20 @@ utcTimestampInIso8601 y mo d h mi s =
 -- Time interval in seconds.
 -- Directly applicable to 'DiffTime' and 'NominalDiffTime'.
 {-# INLINEABLE intervalInSeconds #-}
-intervalInSeconds :: RealFrac seconds => seconds -> TextBuilder
+intervalInSeconds :: (RealFrac seconds) => seconds -> TextBuilder
 intervalInSeconds interval = flip evalState (round interval) $ do
   seconds <- state (swap . flip divMod 60)
   minutes <- state (swap . flip divMod 60)
   hours <- state (swap . flip divMod 24)
   days <- get
-  return $
-    padFromLeft 2 '0' (decimal days) <> ":"
-      <> padFromLeft 2 '0' (decimal hours)
-      <> ":"
-      <> padFromLeft 2 '0' (decimal minutes)
-      <> ":"
-      <> padFromLeft 2 '0' (decimal seconds)
+  return
+    $ padFromLeft 2 '0' (decimal days)
+    <> ":"
+    <> padFromLeft 2 '0' (decimal hours)
+    <> ":"
+    <> padFromLeft 2 '0' (decimal minutes)
+    <> ":"
+    <> padFromLeft 2 '0' (decimal seconds)
 
 -- | Double with a fixed number of decimal places.
 {-# INLINE fixedDouble #-}
@@ -597,7 +622,8 @@ doublePercent decimalPlaces x = fixedDouble decimalPlaces (x * 100) <> "%"
 {-# INLINE hexData #-}
 hexData :: ByteString -> TextBuilder
 hexData =
-  intercalate " " . fmap mconcat
+  intercalate " "
+    . fmap mconcat
     . Split.chunksOf 2
     . fmap byte
     . ByteString.unpack
